/**************************************************************************\
*          Copyright (c) 1992 INRIA Sophia Antipolis, FRANCE.              *
*                                                                          *
* Permission to use, copy, modify, and distribute this material for any    *
* purpose and without fee is hereby granted, provided that the above       *
* copyright notice and this permission notice appear in all copies.        *
* WE MAKE NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS     *
* MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS   *
* OR IMPLIED WARRANTIES.                                                   *
\**************************************************************************/
/**************************************************************************\
*                                                                          *
*  File              : protocol.c                                          *
*  Author            : Thierry Turletti                                    *
*  Last modification : 12/21/92                                            *
*--------------------------------------------------------------------------*
*  Description :  Protocol between stations.                               *
*                                                                          *
\**************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include "protocol.h"



/***************************************************************************\
*                        LOCAL TABLE FUNCTIONS                              *
\***************************************************************************/


void InitTable(t)
     LOCAL_TABLE t;
{
  register int i;

  for(i=0; i<N_MAX_STATIONS; i++){
    bzero(t[i].name, LEN_MAX_NAME);
    t[i].sin_addr = 0;
    t[i].video_encoding = FALSE;
    t[i].video_decoding = FALSE;
    t[i].audio_encoding = FALSE;
    t[i].audio_decoding = FALSE;
    t[i].valid = FALSE;
  }
}


void ResetStation(station)
     STATION *station;
{
  station->valid = FALSE;
  bzero(station->name, LEN_MAX_NAME);
  station->sin_addr = 0;
  station->video_encoding = FALSE;
  station->video_decoding = FALSE;
  station->audio_encoding = FALSE;
  station->audio_decoding = FALSE;
}



void InitStation(p, name, sin_addr, enc_audio, enc_video)
     STATION *p;
     char *name;
     u_long sin_addr;
     BOOLEAN enc_audio, enc_video;
{
  strcpy(p->name, name);
  p->sin_addr = sin_addr;
  p->audio_encoding = enc_audio;
  p->video_encoding = enc_video;
  p->audio_decoding = FALSE;
  p->video_decoding = FALSE;
  p->valid = TRUE;
}


void CopyStation(out, in)
     STATION *out, *in;
{
  
  strcpy(out->name, in->name);
  out->sin_addr = in->sin_addr;
  out->video_encoding = in->video_encoding;
  out->audio_encoding = in->audio_encoding;
  out->video_decoding = in->video_decoding;
  out->audio_decoding = in->audio_decoding;
  out->valid = in->valid;
  out->lasttime = in->lasttime;
}




BOOLEAN ShallIDecode(t, nb_stations, sin_addr)
     AUDIO_TABLE t;
     int nb_stations;
     u_long sin_addr;
{
  register int i;

  for(i=0; i<nb_stations; i++){
    if(t[i].sin_addr == sin_addr){
      if(t[i].audio_decoding)
	return(TRUE);
      break;
    }
  }
  return(FALSE);
}


BOOLEAN AnyAudioDecoding(t, nb_stations)
     AUDIO_TABLE t;
     int nb_stations;
{
  register int i;

  for(i=0; i<nb_stations; i++){
    if(t[i].audio_decoding)
      return(TRUE);
  }
  return(FALSE);
}



void PrintStation(F, p, i)
     FILE *F;
     STATION *p;
     int i;
{
  fprintf(F, "\nSTATION %d ::\n", i);
  fprintf(F, "Name: %s\n", p->name);
  fprintf(F, "audio_encoding : %s\n", (p->audio_encoding ? "Yes" : "No"));
  fprintf(F, "video_encoding : %s\n", (p->video_encoding ? "Yes" : "No"));
  fprintf(F, "valid : %s\n", (p->valid ? "Yes" : "No"));
}


void PrintTable(F, t)
     FILE *F;
     LOCAL_TABLE t;
{
  register int i;

  fprintf(F, "\n\n\t\t*** LOCAL TABLE ***");
  for(i=0; i<N_MAX_STATIONS; i++){
    if(t[i].valid)
      PrintStation(F, &(t[i]), i);
  }
}



/****************************************************************************\
*                  SENDING & RECEIVING CONTROL PACKETS                       *
\****************************************************************************/


int CreateSendSocket(addr, len_addr, port, group, ttl)
     struct sockaddr_in *addr;
     int *len_addr;
     char *port;
     char *group; /* IP address style "225.23.23.6" = "a4.a3.a2.a1" */
     unsigned char *ttl;
{
  int sock;
  unsigned int iport, i1, i2, i3, i4;
  extern BOOLEAN UNICAST;

  if(sscanf(group, "%u.%u.%u.%u", &i4, &i3, &i2, &i1) != 4){
    fprintf(stderr, "CreateSendSocket: Invalid group %s\n", group);
    return(-1);
  }

  iport = atoi(port);
  if((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
    perror("Cannot create datagram socket");
    return(-1);
  }
  *len_addr = sizeof(*addr);
  bzero((char *)addr, *len_addr);
  addr->sin_port = htons((u_short)iport);
  addr->sin_addr.s_addr = htonl((i4<<24) | (i3<<16) | (i2<<8) | i1);
  addr->sin_family = AF_INET;
  if(UNICAST == FALSE){
#ifdef MULTICAST
    setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, ttl, sizeof(*ttl));
#else
    fprintf(stderr, "MULTICAST mode not supported, use only unicast option\n");
    exit(-1);
#endif /* MULTICAST */
  }
  return(sock);
}


int CreateReceiveSocket(addr, len_addr, port, group)
     struct sockaddr_in *addr;
     int *len_addr;
     char *port;
     char *group; /* IP address style "225.23.23.6" = "a4.a3.a2.a1" */
{
  int sock;
  unsigned int iport, i1, i2, i3, i4;
  struct hostent *hp;
  char hostname[64];
  extern BOOLEAN UNICAST;
#ifdef MULTICAST
  int one=1;
  struct ip_mreq imr;
#endif 

  if(!UNICAST){
    if(sscanf(group, "%u.%u.%u.%u", &i4, &i3, &i2, &i1) != 4){
      fprintf(stderr, "CreateReceiveSocket: Invalid group %s\n", group);
      return(-1);
    }
  }

  *len_addr = sizeof(*addr);
  iport = atoi(port);
  if((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
    perror("Cannot create datagram socket");
    exit(-1);
  }
  gethostname(hostname, 64);
  if((hp=gethostbyname(hostname)) == 0){
      fprintf(stderr, "Unknown host %s\n", hostname);
      exit(0);
  }
  if(UNICAST){
    bzero((char *) addr, *len_addr);
    bcopy(hp->h_addr, (char *)addr, sizeof(*addr));
  }else{
#ifdef MULTICAST
    bcopy(hp->h_addr, (char *)&imr.imr_interface.s_addr, hp->h_length);
    imr.imr_multiaddr.s_addr = (i4<<24) | (i3<<16) | (i2<<8) | i1;
    imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
    if(setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &imr,
		  sizeof(struct ip_mreq)) == -1){
      fprintf(stderr, "Cannot join group %s\n", group);
      perror("setsockopt()");
      fprintf(stderr, "Does your kernel support IP multicast extensions ?\n");
      exit(1);
    }
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
#else
    fprintf(stderr, "MULTICAST mode not implemented, use unicast option\n");
    exit(-1);
#endif /* MULTICAST */
  }
  addr->sin_addr.s_addr = INADDR_ANY;
  addr->sin_family = AF_INET;
  addr->sin_port = htons((u_short)iport);
  if (bind(sock, (struct sockaddr *) addr, *len_addr) < 0){
    fprintf(stderr, "Cannot bind socket to %d port\n", iport);
    exit(1);
  }
  return(sock);
}


int SendDeclarePacket(s, addr, len_addr, video_encoding, audio_encoding, name)
     int s;
     struct sockaddr_in *addr;
     int len_addr;
     BOOLEAN video_encoding, audio_encoding;
     char *name;
{
  unsigned char buf[10];
  int len, lw;

  buf[0] = 0xc0;
  if(video_encoding)
    buf[0] |= 2;
  if(audio_encoding)
    buf[0] |= 1;
  strcpy((char *) &(buf[1]), name);
  len = strlen(name) + 1;
  buf[len++] = 0;
  if((lw=sendto(s, (char *)buf, len, 0, (struct sockaddr *)addr, len_addr)) 
     != len){
    perror("Cannot send Declaration Packet on socket");
    return(0);
  }
  return(1);
}



SendAskInfoPacket(s, addr, len_addr)
     int s;
     struct sockaddr_in *addr;
     int len_addr;
{
  unsigned char buf[2];
  int lw;

  buf[0] = 0xc0;
  buf[0] |= (P_ASK_INFO << 2);
  if((lw=sendto(s, (char *)buf, 1, 0, (struct sockaddr *)addr, len_addr)) 
     != 1){
    perror("Cannot send AskInfo Packet on socket");
    return(0);
  }
  return(1);
  
}


SendAbortPacket(s, addr, len_addr)
     int s;
     struct sockaddr_in *addr;
     int len_addr;
{
  unsigned char buf[1];
  int lw;

  buf[0] = 0xc0;
  buf[0] |= (P_ABORT << 2);
  if((lw=sendto(s, (char *)buf, 1, 0, (struct sockaddr *)addr, len_addr)) 
     != 1){
    perror("Cannot send Abort Packet on socket");
    return(0);
  }
  return(1);
  
}


SendFullIntraRequest(host_coder)
     char *host_coder;
{
  struct hostent *hp;
  unsigned long add_long;
  unsigned char buf[1];
  int lw;
  static struct sockaddr_in addr;
  static int sock_report, len_addr;
  static BOOLEAN first_request=TRUE;

  if(first_request){
    bzero((char *)&addr, sizeof(addr));
    if((hp=gethostbyname(host_coder)) == 0){
      if((add_long=inet_addr(host_coder)) == -1){
	fprintf(stderr, "Unknown host : %s\n", host_coder);
	exit(1);
      }else
	addr.sin_addr.s_addr = add_long;
    }else
      bcopy(hp->h_addr, (char *)&addr.sin_addr, hp->h_length);
    addr.sin_port = htons(PORT_VIDEO_CONTROL);
    addr.sin_family = AF_INET;
    len_addr = sizeof(addr);
    if((sock_report = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
      perror("Cannot create datagram socket");
      exit(-1);
    }
    first_request = FALSE;
  }
  buf[0] = P_FULL_INTRA_REQUEST;
  if((lw=sendto(sock_report, (char *)buf, 1, 0, (struct sockaddr *)&addr, 
		len_addr)) != 1){
    perror("Cannot send Full Intra Request Packet on socket");
    return(0);
  }
  return(1);
  
}


SendInfoPacketLost(host_coder, from_gob, to_gob, image_number, cif_number)
     char *host_coder;
     int from_gob, to_gob, image_number, cif_number;
{
  struct hostent *hp;
  unsigned long add_long;
  unsigned char buf[3];
  int lw;
  static struct sockaddr_in addr;
  static int sock_report, len_addr;
  static BOOLEAN first_info=TRUE;

  if(first_info){
    bzero((char *)&addr, sizeof(addr));
    if((hp=gethostbyname(host_coder)) == 0){
      if((add_long=inet_addr(host_coder)) == -1){
	fprintf(stderr, "Unknown host : %s\n", host_coder);
	exit(1);
      }else
	addr.sin_addr.s_addr = add_long;
    }else
      bcopy(hp->h_addr, (char *)&addr.sin_addr, hp->h_length);
    addr.sin_port = htons(PORT_VIDEO_CONTROL);
    addr.sin_family = AF_INET;
    len_addr = sizeof(addr);
    if((sock_report = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
      perror("Cannot create datagram socket");
      exit(-1);
    }
    first_info = FALSE;
  }
  buf[0] = P_SEND_INFO_PACKET_LOST | (cif_number << 4);
  buf[1] = (from_gob << 4) | (to_gob & 0x0f);
  buf[2] = image_number & 0xff;
  if((lw=sendto(sock_report, (char *)buf, 3, 0, (struct sockaddr *)&addr, 
		len_addr)) != 3){
    perror("Cannot send Send Info Packet Lost on socket");
    return(0);
  }
  return(1);
}



/****************************************************************************\
*                               PIPE  UTILITIES                              *
\****************************************************************************/


BOOLEAN HandleCommands(fd, size, rate_control, port_vfc, rate_max, FREEZE,
		       LOCAL_DISPLAY)
     int fd[2];
     int *size;
     int *rate_control;
     int *port_vfc;
     int *rate_max;
     BOOLEAN *FREEZE;
     BOOLEAN *LOCAL_DISPLAY;
{
  struct timeval tim_out;
  int mask;
  u_char command[500];
  int lr, i=0;

  mask = 1<<fd[0];
  tim_out.tv_sec = 0;
  tim_out.tv_usec = 0;
  if(select(fd[0]+1, (fd_set *)&mask, (fd_set *)0, (fd_set *)0, 
	    &tim_out) != 0){
    if((lr=read(fd[0], command, 500)) <= 0){
      perror("HandleCommands'read ");
      return FALSE;
    }
    do{
      switch(command[i]){
      case PIP_QUIT: /* One byte message */
	return FALSE;
      case PIP_SIZE: /* Two bytes message */
	*size = (int)command[++i];
	i ++;
	break;
      case PIP_FREEZE: /* One byte message */
	*FREEZE = TRUE;
	i ++;
	break;
      case PIP_UNFREEZE: /* One byte message */
	*FREEZE = FALSE;
	i ++;
	break;
      case PIP_WCONTROL: /* One byte message */
	*rate_control=WITHOUT_CONTROL;
	i ++;
	break;
      case PIP_RCONTROL: /* One byte message */
	*rate_control = REFRESH_CONTROL;
	i ++;
	break;
      case PIP_QCONTROL: /* One byte message */
	*rate_control = QUALITY_CONTROL;
	i ++;
	break;
      case PIP_VFC_PORT: /* Two bytes message */
	*port_vfc = (int)command[++i];
	i ++;
	break;
      case PIP_MAX_RATE: /* Five bytes message */
	sscanf((char *) &command[++i], "%d", rate_max);
	i += 4;
	break;
      case PIP_LDISP: /* One byte message */
	*LOCAL_DISPLAY = TRUE;
	i ++;
	break;
      case PIP_NOLDISP: /* One byte message */
	*LOCAL_DISPLAY = FALSE;
	i ++;
	break;
      }
    }while(lr>i);
  }
  return(TRUE);
}
      


int ItIsTheEnd(fd)
     int fd[2];
  {
    struct timeval tim_out;
    int mask;

    mask = 1<<fd[0];
    tim_out.tv_sec = 0;
    tim_out.tv_usec = 0;
    if(select(fd[0]+1, (fd_set *)&mask, (fd_set *)0, (fd_set *)0, 
	      &tim_out) != 0)
      return(1);
    else
      return(0);
  }
    



SendPP_QUIT(fd)
     int fd[2];
{
  u_char buf=PP_QUIT;

  if(write(fd[1], (char *)&buf, 1) != 1){
#ifdef DEBUG
    perror("SendPP_QUIT'write");
#endif
  }
}



BOOLEAN SendPP_FORK(fd)
     int fd[2];
{
  u_char buf=PP_FORK;

  if(write(fd[1], (char *)&buf, 1) != 1){
    perror("SendPP_UIT'write");
    return(FALSE);
  }
  return(TRUE);
}



SendPP_NEW_FORMAT(fd, format)
     int fd[2];
{
  u_char buf[2];

  buf[0]= (u_char) PP_NEW_FORMAT;
  buf[1] = (u_char) format;

  if(write(fd[1], (char *)buf, 2) != 2){
#ifdef DEBUG
    perror("SendPP_NEW_FORMAT'write");
#endif
  }
}



SendPP_NEW_TABLE(t, nb_stations, fd)
     LOCAL_TABLE t;
     int nb_stations; /* less than 255 .. */
     int fd[2];
{
  u_char *buf;
  int lbuf, lw;
  register int i;
  AUDIO_TABLE t_audio;

  lbuf = nb_stations*sizeof(AUDIO_STATION);
  buf = (u_char *)malloc(lbuf+2);
  buf[0] = (u_char) PP_NEW_TABLE;
  buf[1] = (u_char) nb_stations;
  for(i=0; i<nb_stations; i++){
    t_audio[i].sin_addr = t[i].sin_addr;
    t_audio[i].audio_decoding = t[i].audio_decoding;
  }
  bcopy((char *)t_audio, (char *)&buf[2], lbuf);
  if((lw=write(fd[1], (char *)buf, lbuf+2)) < 0){
    perror("SendPP_NEW_TABLE'write");
  }
  free(buf);
}


int GetType(buf)
     u_char *buf;
{
  return(buf[0] >> 6);
}


int GetSubType(buf)
     u_char *buf;
{
  return((buf[0] >> 2) & 0x0f);
}


BOOLEAN ChangeAudioDecoding(t, name, audio_decoding)
     LOCAL_TABLE t;
     char *name;
     BOOLEAN audio_decoding;
{
  register int i;

  for(i=0; i<N_MAX_STATIONS; i++){
    if((strcmp(name, t[i].name) == 0) && t[i].valid){
      if(t[i].audio_decoding != audio_decoding){
	t[i].audio_decoding = audio_decoding;
	return(TRUE);
      }else
	break;
    }
  }
  return(FALSE);
}


BOOLEAN ChangeVideoDecoding(t, name, video_decoding)
     LOCAL_TABLE t;
     char *name;
     BOOLEAN video_decoding;
{
  register int i;

  for(i=0; i<N_MAX_STATIONS; i++){
    if((strcmp(name, t[i].name) == 0) && (t[i].valid)){
      if(t[i].video_decoding != video_decoding){
	t[i].video_decoding = video_decoding;
	return(TRUE);
      }else
	break;
    }
  }
  return(FALSE);
}


/****************************************************************************\
*                         IVS DAEMON  UTILITIES                              *
\****************************************************************************/


int SendTalkRequest(sock, name)
     int sock;
     char *name;
{
  int lwrite, len;
  char data[10];

  len = strlen(name) + 1;
  bzero(data, 10);
  data[0] = (u_char) CALL_REQUEST;
  strcpy(&data[1], name);
  if((lwrite = write(sock, data, len)) != len)
    return(FALSE);
  else
    return(TRUE);
}
     


int SendTalkAbort(sock)
     int sock;
{
  int lwrite;
  char data;

  data = (u_char) CALL_ABORT;
  if((lwrite = write(sock, &data, 1)) != 1)
    return(FALSE);
  else
    return(TRUE);
}
     


int SendTalkAccept(sock)
     int sock;
{
  int lwrite;
  char data;

  data = (u_char) CALL_ACCEPTED;
  if((lwrite = write(sock, &data, 1)) != 1)
    return(FALSE);
  else
    return(TRUE);
}
     


int SendTalkRefuse(sock)
     int sock;
{
  int lwrite;
  char data;

  data = (u_char) CALL_REFUSED;
  if((lwrite = write(sock, &data, 1)) != 1)
    return(FALSE);
  else
    return(TRUE);
}
     


